local TR = TICRATE
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			return freeslot(item)
		end
	end
end
local numtotrans = {
	[10] = FF_TRANS90,
	[9] = FF_TRANS90,
	[8] = FF_TRANS80,
	[7] = FF_TRANS70,
	[6] = FF_TRANS60,
	[5] = FF_TRANS50,
	[4] = FF_TRANS40,
	[3] = FF_TRANS30,
	[2] = FF_TRANS20,
	[1] = FF_TRANS10,
	[0] = 0,
}

SafeFreeslot("SPR_SBSM")
sfxinfo[SafeFreeslot("sfx_subsma")].caption = "Mine activate"
sfxinfo[SafeFreeslot("sfx_subsme")] = {
	caption = "Explosion",
	flags = SF_X8AWAYSOUND
}
sfxinfo[SafeFreeslot("sfx_subsmt")].caption = "/"

states[SafeFreeslot("S_SUBSPACE_MINE")] = {
	sprite = SPR_SBSM,
	frame = A,
	action = function(mine)
		mine.activatein = 2
		mine.activatewait = 0
		mine.primed = false
		mine.fade = 0
	end,
	tics = -1
}

mobjinfo[SafeFreeslot("MT_SUBSPACE_MINE")] = {
	doomednum = -1,
	spawnstate = S_SUBSPACE_MINE,
	deathsound = sfx_subsme,
	activesound = sfx_subsma,
	seesound = sfx_subsmt,
	spawnhealth = 1,
	height = 32*FRACUNIT,
	radius = 16*FRACUNIT,
	flags = MF_SOLID|MF_RUNSPAWNFUNC
}

local function GetActorZ(actor,targ,type)
	if type == nil then type = 1 end
	if not (actor and actor.valid) then return 0 end
	if not (targ and targ.valid) then return 0 end
	
	local flip = P_MobjFlip(actor)
	
	--get z
	if type == 1
		if flip == 1
			return actor.z
		else
			return actor.z+actor.height-targ.height
		end
	--get top z
	elseif type == 2
		if flip == 1
			return actor.z+actor.height
		else
			return actor.z-targ.height
		end
		
	--RELATIVE
	--get z
	--since theres a good chance we're using these with P_SpawnMobjFromMobj,
	--make the distances absolute (FU, not mo.scale) since ^ already
	--does that
	elseif type == 3
		if flip == 1
			--ez
			return 0
		else
			return FixedDiv(actor.height-targ.height,actor.scale)
		end
	--get top z
	elseif type == 4
		if flip == 1
			return FixedDiv(actor.height,actor.scale)
		else
			return -FixedDiv(targ.height,actor.scale)
		end
	end
	
	return 0
end

local function TakisGetCameraMobj()
	local cam = camera
	if (displayplayer and displayplayer.valid)
		
		if not CV_FindVar("chasecam").value
			
			local pmo = displayplayer.realmo
			if not (pmo and pmo.valid)
				pmo = displayplayer.mo
			end
			
			if (pmo and pmo.valid)
				local th = P_SpawnMobj(
					pmo.x,
					pmo.y,
					GetActorZ(pmo,pmo,2),
					MT_RAY
				)
				th.angle = pmo.angle
				pmo = th
			end
			cam = pmo
		else
			cam = camera
		end
		
		if displayplayer.awayviewtics
		and (displayplayer.awayviewmobj and displayplayer.awayviewmobj.valid)
			cam = displayplayer.awayviewmobj
		end
		
		if not (cam and cam.valid)
			cam = camera
		end
	end
	return cam
end

local function delete3d(door)
	if not door.made3d
		door.flags2 = $ &~MF2_DONTDRAW
		return
	end
	
	if door.sides
		for k,v in pairs(door.sides)
			if v and v.valid
				P_RemoveMobj(v)
			end
		end
	end
	
	door.made3d = false
	door.flags2 = $ &~MF2_DONTDRAW
end

local TAKIS_3D_FLAGS = MF_NOGRAVITY|MF_NOCLIPTHING|MF_NOCLIP|MF_SCENERY
local function ThreeDThinker(door)
	local dist = 0
	local cullout = true
	local doculling = true
	if doculling
		local cam = TakisGetCameraMobj()
		
		dist = R_PointToDist2(cam.x,cam.y, door.x,door.y)
		
		local thok = P_SpawnMobj(cam.x, cam.y, cam.z, MT_RAY)
		thok.angle = cam.angle
		thok.flags2 = $|MF2_DONTDRAW
		if dist <= 5000*FU
		and P_CheckSight(thok,door)
			cullout = false
		end
		
		if not cullout
			local back = FixedAngle(AngleFixed(thok.angle)+180*FU)
			local diff = FixedAngle(AngleFixed(R_PointToAngle2(thok.x, thok.y, door.x, door.y))-AngleFixed(back))
			if AngleFixed(diff) > 180*FU
				diff = InvAngle(diff)
			end
			
			--in the cameras view
			if AngleFixed(diff) > 90*FU
				cullout = false
			else
				cullout = true
			end
		end
		
		if not door.health
			cullout = true
		end
		P_RemoveMobj(thok)
	end
	
	if cullout
		delete3d(door)
		return
	end
	
	local trans = numtotrans[door.fade]
	
	if not cullout

		if not door.made3d
			local list
			local flip = P_MobjFlip(door)
			door.flags2 = $|MF2_DONTDRAW
			
			door.sides = {}
			list = door.sides
			
			for i = 1,4
				local angle = door.angle+(FixedAngle(90*FU*(i-1)))
				list[0+i] = P_SpawnMobjFromMobj(door,
					P_ReturnThrustX(nil,angle,16*FU),
					P_ReturnThrustY(nil,angle,16*FU),
					0,MT_RAY
				)
				list[0+i].frame = A|FF_FULLBRIGHT
				list[0+i].sprite = SPR_SBSM
				list[0+i].tics,list[0+i].fuse = -1,-1
				list[0+i].flags = TAKIS_3D_FLAGS
				list[0+i].renderflags = $|RF_PAPERSPRITE|RF_NOSPLATBILLBOARD
				list[0+i].angle = angle+ANGLE_90
				list[0+i].height = 32*FU
				list[0+i].radius = 0
				list[0+i].destscale = door.scale
				list[0+i].scalespeed = list[0+i].destscale + 1
				P_SetOrigin(list[0+i],
					list[0+i].x,
					list[0+i].y,
					GetActorZ(door,list[0+i],1)
				)
			end
			list[5] = P_SpawnMobjFromMobj(door,0,0,0,MT_RAY)
			list[5].frame = B|FF_FULLBRIGHT
			list[5].sprite = SPR_SBSM
			list[5].tics,list[5].fuse = -1,-1
			list[5].flags = TAKIS_3D_FLAGS
			list[5].renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
			list[5].angle = door.angle
			list[5].height = 0
			list[5].scale = door.scale
			list[5].destscale = door.destscale
			list[5].scalespeed = door.scalespeed
			P_SetOrigin(list[5],list[5].x,list[5].y,GetActorZ(door,list[5],2))
			
			list[6] = P_SpawnMobjFromMobj(door,0,0,0,MT_RAY)
			list[6].frame = B|FF_FULLBRIGHT
			list[6].sprite = SPR_SBSM
			list[6].tics,list[5].fuse = -1,-1
			list[6].flags = TAKIS_3D_FLAGS
			list[6].renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
			list[6].angle = door.angle
			list[6].height = 0
			list[6].scale = door.scale
			list[6].destscale = door.destscale
			list[6].scalespeed = door.scalespeed
			P_SetOrigin(list[6],list[6].x,list[6].y,GetActorZ(door,list[6],1))
			
			list[7] = P_SpawnMobjFromMobj(door,0,0,0,MT_RAY)
			list[7].frame = C|FF_FULLBRIGHT
			list[7].sprite = SPR_SBSM
			list[7].tics,list[5].fuse = -1,-1
			list[7].flags = TAKIS_3D_FLAGS
			list[7].angle = door.angle
			list[7].height = door.height
			list[7].scale = door.scale
			list[7].destscale = door.destscale
			list[7].scalespeed = door.scalespeed
			P_SetOrigin(list[7],list[7].x,list[7].y,GetActorZ(door,list[7],1))
			
			door.made3d = true
		
		--update positions
		else
			local list = door.sides
			
			for i = 1,4
				local angle = door.angle+(FixedAngle(90*FU*(i-1)))
				list[0+i].angle = angle+ANGLE_90
				list[0+i].height = 32*FU
				list[0+i].radius = 0
				list[0+i].destscale = door.scale
				list[0+i].scalespeed = list[0+i].destscale + 1
				list[0+i].frame = ($ &~FF_TRANSMASK)|trans
				P_MoveOrigin(list[0+i],
					door.x+P_ReturnThrustX(nil,angle,16*door.scale) + door.momx,
					door.y+P_ReturnThrustY(nil,angle,16*door.scale) + door.momy,
					GetActorZ(door,list[0+i],1) + door.momz
				)
				
				if door.fade == 10
					list[0+i].flags2 = $|MF2_DONTDRAW
				end
			end
			
			--TODO: these splats are positioned incorrectly on fofs
			list[5].angle = door.angle
			list[5].height = 0
			list[5].scale = door.scale
			list[5].shadowscale = door.scale*14/10
			list[5].frame = ($ &~FF_TRANSMASK)|trans
			P_MoveOrigin(list[5],
				door.x + door.momx,
				door.y + door.momy,
				(P_MobjFlip(door) == 1 and door.z + door.height or door.z) + door.momz
			)
			if door.fade == 10
				list[5].flags2 = $|MF2_DONTDRAW
			end
			
			--P_SetOrigin(list[6],door.x,door.y,door.z)
			list[6].angle = door.angle
			list[6].height = 0
			list[6].scale = door.scale
			list[6].frame = ($ &~FF_TRANSMASK)|trans
			P_MoveOrigin(list[6],
				door.x + door.momx,
				door.y + door.momy,
				(P_MobjFlip(door) == 1 and door.z or door.z + door.height) + door.momz
			)
			if door.fade == 10
				list[6].flags2 = $|MF2_DONTDRAW
			end
			
			list[7].angle = door.angle
			list[7].scale = door.scale
			list[7].frame = ($ &~FF_TRANSMASK)|trans
			P_MoveOrigin(list[7],
				door.x + door.momx,
				door.y + door.momy,
				(P_MobjFlip(door) == 1 and door.z or door.z + door.height) + door.momz
			)
			if door.fade == 10
				list[7].flags2 = $|MF2_DONTDRAW
			end
			
		end
	end
	
end

addHook("MobjThinker",function(mine)
	if not (mine and mine.valid) then return end
	
	ThreeDThinker(mine)
	P_ButteredSlope(mine)
	
	if R_PointToDist2(0,0,mine.momx,mine.momy) <= 3*mine.scale
		if mine.activatein
			if not mine.activatewait
				S_StartSound(mine,mine.info.activesound)
				mine.activatewait = TR*3/2
				mine.activatein = $-1
				
				if mine.activatein == 0
					S_StartSound(mine,mine.info.seesound)
				end
			else
				mine.activatewait = $-1
			end
			
		end
		
	end
	
	if not mine.activatein
		if not mine.primed
			mine.flags = $|MF_SPECIAL
		end
		mine.primed = true
		
		if mine.fade < 10
		and (leveltime & 1)
			mine.fade = $+1
		end
	end
	
end,MT_SUBSPACE_MINE)

local function SpawnSparks(mo)
	for i = 10,P_RandomRange(15,20)
		local spark = P_SpawnMobjFromMobj(mo,
			P_RandomRange(-32,32)*FU,
			P_RandomRange(-32,32)*FU,
			FixedDiv(mo.height/2,mo.scale) + P_RandomRange(-16,16)*FU,
			MT_THOK
		)
		spark.state = S_SPRK1
		spark.fuse = 16
		spark.colorized = true
		spark.blendmode = AST_ADD
		spark.color = SKINCOLOR_MAJESTY
		
		P_Thrust(spark,
			R_PointToAngle2(spark.x,spark.y, mo.x,mo.y),
			-3*spark.scale
		)
		P_SetObjectMomZ(spark,P_RandomRange(-3,3)*FU)
	end
end

local function SetPurplePlanes(mo)
	local fakerange = 256*mo.scale
	searchBlockmap("lines", function(ref, line)
		if line.frontside
			line.frontside.toptexture = R_TextureNumForName("VIOWALL")
			line.frontside.bottomtexture = line.frontside.toptexture
		end
		if line.backside
			line.backside.toptexture = R_TextureNumForName("VIOWALL")
			line.backside.bottomtexture = line.backside.toptexture
		--there is void behind this line
		else
			line.frontside.midtexture = R_TextureNumForName("VIOWALL")
		end
		
		if line.frontsector
			if line.frontsector.floorpic ~= "F_SKY1"
				line.frontsector.floorpic = "VIOWALL"
			end
			if line.frontsector.ceilingpic ~= "F_SKY1"
				line.frontsector.ceilingpic = "VIOWALL"
			end
		end
		if line.backsector
			if line.backsector.floorpic ~= "F_SKY1"
				line.backsector.floorpic = "VIOWALL"
			end
			if line.backsector.ceilingpic ~= "F_SKY1"
				line.backsector.ceilingpic = "VIOWALL"
			end
		end
	end, 
	mo,
	mo.x-fakerange, mo.x+fakerange,
	mo.y-fakerange, mo.y+fakerange)
	
end

addHook("TouchSpecial",function(mine,me)
	if not (mine and mine.valid) then return end
	if not (me and me.valid) then return end
	if not (me.health) then return end
	
	local p = me.player
	
	local sfx = P_SpawnGhostMobj(mine)
	sfx.flags2 = $|MF2_DONTDRAW
	sfx.fuse = 12*TR
	
	S_StartSound(sfx,mine.info.deathsound)
	S_StartSound(sfx,mine.info.deathsound)
	S_StartSound(sfx,mine.info.deathsound)
	
	P_KillMobj(me,mine,mine,DMG_INSTAKILL)
	me.color = SKINCOLOR_MAJESTY
	me.colorized = true
	S_StartSound(me,mine.info.deathsound)
	
	SpawnSparks(mine)
	SpawnSparks(me)
	SetPurplePlanes(mine)
	
	delete3d(mine)
end,MT_SUBSPACE_MINE)

addHook("MobjRemoved",delete3d,MT_SUBSPACE_MINE)

COM_AddCommand("spawntripmine",function(p)
	if gamestate ~= GS_LEVEL
		prn(p,"You can't use this right now.")
		return
	end
	
	local me = p.realmo
	if not (me and me.valid)
		prn(p,"You can't use this right now.")
		return
	end
	
	local mine = P_SpawnMobjFromMobj(me,
		P_ReturnThrustX(nil,me.angle,64*FU),
		P_ReturnThrustY(nil,me.angle,64*FU),
		0,
		MT_SUBSPACE_MINE
	)
	mine.angle = me.angle
	mine.scale = me.scale

end,COM_ADMIN)